REBOL [
	Title: "REBOL/Services - Admin Service Dialect"
	Author: "Carl Sassenrath"
	Notice: "Confidential source code. Under development."
	Copyright: "REBOL Technology 2005-2008, All rights reserved."
]

add-default-service [
	name: 'admin
	title: "Administrative Service"
	description: {
		Admin service implements updating server files, downloading logs,
		changing configuration, managing users, and more.
	}

	; locals (remember ARGS and RESULT are already local to the rules!)
	file: fp: size: user: user-name: pass: tags: info: expire: none

	commands: [

		#need admin ; all commands require admin permission

			'info #doc [
				en "General info about this service."
				it "Informazioni generiche riguardo questo servizio."
			]
			(result: reduce [base-dir])
		|
			'debug #doc [
				en "Control debugging mode for service."
				it "Controlla la modalit di debug per questo servizio."
			]
		|
			'restart #doc [
				en "Stop and restart the server."
				it "Riavvia il server."
			]
			(stop-server session true)
		|
			'stop #doc [
				en "Stop the server. Cannot be restarted remotely."
				it "Ferma il server. Non  possibile riavviarlo da remoto."
			]
			(debug system ['stop 'admin-user session/user-id])
			; Note: command evaluator detects STOP command itself.
			(stop-server session false)
		|
			'config #doc [
				en "Get or set the server configuration."
				it "Ottieni o modifica la configurazione del server."
			] [
				(result: 0)
				"Set specific config values."
				into [any [word! value]]
				|
				"Get specific config values."
				some word!
				|
				"Get all config values."
				end (result: form session/config) ; 2.6.0
			]
		|
			'log #doc [
				en "Log file access and control."
			] [
				(file: join base-dir session/config/log-file)
				'info (
					result: compose [
						file: (session/config/log-file)
						watch: (session/config/log-watch)
						size: (attempt [size? file])
						date: (attempt [modified? file])
					]
				)
				| 'purge (result: attempt [delete file true])
				| 'config (result: none)
				| [set size integer! | 'all | none (size: 8000)] (
					if fp: attempt [open/binary/seek file] [ ; 2.6.2
						if not integer? size [size: length? fp]
						result: copy/part at max 0 (length? fp) - size fp size
						; Remove possible partial line (from seek):
						if size <> length? fp [
							remove/part result any [find/tail result newline 0]
						]
						close fp
					]
				)
			]
		|
			'files [
				(result: copy [])
				'put "Update one or more modules of the server." [
					some [args: file! binary! (
						repend result [args/1 attempt [write join base-dir args/1 args/2 true]]
					)]
				]
				| 'get "Get one or more modules of the server." [
					some [set args file! (
						repend result [args attempt [read join base-dir args]]
					)]
				]
				| 'list "List the modules of the server."
					(result: load base-dir)
			]

	;-- User Admin:
		|
			'user-info #doc [
				en "Summarize information about user base."
				it "Riassumi informazioni sulla base utenti."
			]
			result: to end ; #users, #active, #online, last changed
		|
			'users #doc [
				en "Return all attributes of all users as blocks."
			] (result: list-users none)
		|
			'user-list #doc [
				en "Return specific attributes of one or more users."
				it "Ritorna specifici attributi di uno o pi utenti."
			] args: some word! (result: list-users args) ; Multiple users doesn't work
		|
			'user-add #doc [
				en "Add one or more new users."
				it "Aggiungi uno o pi utenti."
			] [
				some [
					into [
						set user string!
						set user-name string!
						set pass [string! | binary!]
						set tags opt block!
						set info opt binary!
						set expire opt [date! | integer!] ; (#days)
					]
					(create-user/tags user user-name pass any [tags []])
				] (save-userbase result: true)
			]
		|
			'user-delete #doc [
				en "Delete one or more existing users (by name or id#)."
				it "Cancella uno o pi utenti esistenti (per nome o ID)."
			] [
				any [set user [string! | integer!]]
			] (delete-user user)
		|
			'user-pass #doc [
				en "Change a user's password."
				it "Cambia la password di un utente."
			] [
				args: integer! string!
				(
					change-pass args/1 args/2
					save-userbase
					result: true
				)
			]
		|
			'user-allow #doc [
				en "Set user permission tags."
				it "Imposta i permessi di uno o pi utenti."
			] [
				some [user integer! tags block!]
			]
		|
			'user-change #doc [
				en "Change other user attributes."
				it "Modifica altri attributi di uno o pi utenti."
			] [
				args: integer! some [word! skip]
				(
					change-user args/1 next args
					save-userbase
					result: true
				)
		]
	]

	stop-server: func [session restart] [
		debug system ['restart 'admin-user session/user-id]
		if not find session/config/flags 'transient [
			attempt [close-all-sessions session/config] ; !!!
			; how to close main server port!!! ? config?
			attempt [close serv-port]
			either restart [
				launch/quit system/options/script ; does not return
			][
				quit
			]
		]
	]
]
