REBOL [
	Title: "REBOL/Services - Session Management"
	Author: "Carl Sassenrath"
	Notice: "Confidential source code. Under development."
	Copyright: "REBOL Technology 2005-2008, All rights reserved."
]

;-- Session Management -------------------------------------------------------
;
;   Session object - Used by server and client.
;
;   This object and its code keep track of user session information,
;   such as buffers, requests, encryption, etc.  When a user connects
;   a new session object is created, and it serves as the unique
;   identification (the user-id is not enough, because the same user
;   may be connected from more than one location.)
;
;   Session types:
;
;   PERSISTENT - The connection is always open. This is common for
;   direct-TCP services. The session object is attached to the
;   connection port, but the packet SID must also validate that.
;   The session life is the same as the connection time. However,
;   the server is allowed to terminate sessions due to inactivity,
;   exceeding command limits, permission violation, etc.
;
;   TRANSIENT - The connection is temporary. This is common for CGI
;   services. The connection is made, the command sent, the reply
;   is returned, and the connection is closed. For these, the SID
;   is critical to find the relevant session info, such as the
;   encryption key, user, request sequence number, etc. If no SID
;   is provided, then the service must be either OPEN or the user
;   must be sending the LOGON command. Transient sessions are
;   terminated by a LOGOFF command, a session timeout, or error.

session-obj: context [
	id:             ; session id number
	seq:            ; current packet sequence number
	config:         ; service or client config object
	flags:          ; block of session flags, eg: [open server]
	changed:        ; block of change flags, eg: [login logout]
	remote-ip:      ; IP address of client (as seen by server)

	port:           ; reference back to this session's port
	recv-buf:       ; the receive buffer
	port-timeout:   ; timeout in seconds

	send-queue:     ; queue of items waiting to be sent
	sent-queue:     ; queue of items already sent
	callback:       ; optional callback on request

	header:         ; the current packet header (none if not yet found)
	request:        ; the current command request
	results:        ; the command results
	this-service:   ; current service context

	max-header:     ; Maximum header. Allows for trash at start.
	max-request:    ; Maximum request size. Abort if greater.
	max-payload:    ; Maximum payload size. Abort if greater.

	user-id:        ; user id (0 until successful login)
	encode:         ; packet encoding method
	challenge:      ; authentication challenge (SHA1)
	crypt-key:      ; session encryption key (cloak, AES)
	crypt-len:      ; session encryption length (bits) eg: 128, 72, 56
	rsa-key:        ; RSA key
	encrypt-port:   ; port used for encryption (AES)
	decrypt-port:   ; port used for decryption (AES)
	language:       ; client's natural language

	start-date:     ; when session started
	access-date:    ; when last access was made
	login-date:     ; when login occurred (NOTE: indicates authenticated)
		none
]

; Fields that are saved in a file on transient servers:
saved-fields: [
	id seq remote-ip user-id flags changed language
	encode challenge crypt-key crypt-len rsa-key
	start-date access-date login-date 
]

make-session: func [
	"Create and return a new session object."
	port [port! none!]
	conf [object!] "Server/client configuration object"
	/local session id
][
	session: make session-obj [
		id: 0  ; no session
		seq: 0
		config: conf
		flags: copy []
		changed: copy []
		recv-buf: make binary! 10000
		results: copy []
		send-queue: copy []
		sent-queue: copy []
		start-date: access-date: now
	]
	session/port: port
	reset-session session
	session/callback: get in conf 'callback
	if port [port/user-data: session]
	; If we are a server, give session an id and add it to the list:
	if find conf/flags 'server [
		; Add session to the session list:
		until [ ; find an available session number
			id: conf/session-id + 1
			if id > 999999 [id: 1]
			not find conf/session-list id
		]
		conf/session-id: id
		session/id: id
		insert conf/session-list reduce [id session]
		debug session ['made id 'of (length? conf/session-list) / 2]
	]
	session
]

reset-session: func [
	"Reset session to non-authenticated state."
	session [object!]
][
	debug session ['reset-session session/id]
	clear session/changed
	; Copy various fields from the config:
	foreach word [
		port-timeout encode rsa-key
		max-header max-request max-payload
	][set in session word get in session/config word]
	; Reset authentication fields:
	session/user-id: 0
	session/challenge: none
	session/login-date: none ; indicates authenticated ok
	; Reset encryption (initial key is public access):
	session/crypt-key: session/config/open-key
	session/crypt-len: 0
	close-crypt-ports session
]

free-session: func [
	"Free a session and its resources."
	session [object! none!]
][
	if not object? session [exit]
	debug session ['free-session session/id]
	close-crypt-ports session
	remove/part find session/config/session-list session/id 2
	; Give GC a little help:
	foreach word next first session [set in session word none]
]

close-all-sessions: func [
	"Close all current sessions (used for restart)"
	config [object!]
	/local list
][
	list: extract/index config/session-list 2 2 ; copies it too
	foreach sess list [
		attempt [close sess/port]
		free-session sess
	] 
]

session-service?: func [
	"Return the service session object for a port."
	port [port!]
][
	port/user-data
]

sweep-sessions: func [
	"Garbage collect older unused sessions."
	config
	/local stamp list when
][
	stamp: now
	either find config/flags 'transient [
		list: load config/session-dir
		debug session ['sweeping length? list]
		foreach file list [
			file: to-session-file config file
			when: modified? file
			if (difference stamp when) > config/session-timeout [
				attempt [delete file]
			]
		]
	][
		; (direct)
		debug session ['sweeping (length? config/session-list) / 2]
		remove-each [id sess] config/session-list [
			(difference stamp sess/access-date) > config/session-timeout
		]
	]
]

load-session-file: func [
	"Load session from a file (required for transient servers like CGI)"
	config
	id [integer!]
	/local session
][
	attempt [
		session: load to-session-file config id
		session: construct/with session session-obj
		session/config: config
		if session/id = id [
			repend session/config/session-list [id session] ; need for valid-session
			return session
		]
	] ; otherwise returns NONE
]

save-session-file: func [
	"Save session to a file (required for transient servers like CGI)"
	config
	session [object!]
	/local out
][
	out: make string! 1000
	foreach word saved-fields [
		repend out [word ": " mold session/:word newline]
	]
	write to-session-file config session/id out
]

to-session-file: func [
	config
	id
	/local file
][
	file: to-file id
	insert/dup file #"0" 7 - length? file
	head insert file config/session-dir
]

valid-session?: func [
	"Validate the session id found in the packet header."
	session [object!]
	/local id seq
][
	; In the case of direct connections, the packet's session id had
	; better match our port's session id, otherwise it's a violation.
	; For non-persistent requests like CGI, the packet's session tells
	; us how to find our session info on the server (needed to decrypt)/
	; Return TRUE for valid session, otherwise  FALSE will abort the packet.
	id: ph-session? session/header
	debug session ['valid-session? 'header-id id 'sess-id session/id]

	either not find session/config/flags 'server [
		; CLIENT:
		; If this is our first reply, store the session id,
		; otherwise confirm that it is what we think it should be.
		if any [
			session/id = 0
			session/id = id
		][
			session/id: id
			return true
		]
		return false
	][
		; SERVER:
		; Get sequence # if it's been provided:
		all [
			block? session/request
			block? seq: session/request/1
			seq: select seq 'seq
		]
		seq: any [seq 1]
		debug session ['sequence 'session-1 session/seq 'packet seq]

		; If it is a new session, the sender does not yet
		; know the session id, so it will be zero, and that's ok.
		; But, we must reset the session to prevent a hack that would
		; bypass sequence checking on an authenticated session.
		if not positive? id [
			reset-session session
			session/seq: seq
			return true
		]

		session/access-date: now ; update session timestamp

		; Verify that the session sent to us matches what we think
		; it should be and that the session is still in the cache.
		if all [
			session/id = id
			find session/config/session-list id
		][
			; If user is not logged-in, we don't need to verify seq:
			either session/user-id = 0 [
				session/seq: seq
				return true
			][
				; Verify that the sequence number within the request block
				; is what we think it should be, otherwise something is wrong
				; (or a man-in-the-middle attack is underway).
				session/seq: session/seq + 1
				if session/seq = seq [return true]
			]
		]
		; Invalid session id or sequence error, that's not allowed:
		append session/changed 'reset
		return false
	]
]

login-session: func [
	"[Server] Authenticate user, return challenge, else word if error."
	session [object!]
	user-name [string!]
	crypt-len [integer! none!]
	/local user
][
	; Start the login procedure. Verify that the user exists and
	; make the session challenge. Send a reply to the user that includes
	; the challenge, user-salt (for storing encoded passwords on client),
	; encoding method, and key length (allows reduced strength when
	; server or client is running export-restricted). After reply is
	; sent, the new session encryption key is generated. The next
	; client packet must be properly encrypted, or the login will
	; fail. The login-date field indicates a successful login.
	; (See the crypt-session and confirm-session functions below.)
	user: find-user user-name
	; !!! in config: allow login attempt to return login
	; even if the user does not exist - prevents username
	; snooping.
	if not user [
		debug login ['no-user user-name]
		return 'no-user
	]
	if not able-user? user [
		debug login ['bad-user user-name first user]
			return 'bad-user
	]
	debug login ['challenge user-name first user]
	session/user-id: first user
	if binary? session/config/priv-key [
		session/crypt-len: min crypt-len (8 * length? session/config/priv-key)
	]
	append session/changed 'login
	session/challenge: repend make-challenge session/user-id [
		pe-sess-cloak session/crypt-len ; !!! how to know encoding method?
	]
	; returns [challenge salt encoding crypt-len]
]

logout-session: func [
	"Terminate the authenticated, encrypted session."
	session [object!]
][
	; Delay until the logout reply has been handled:
	append session/changed 'reset
]

crypt-session: func [
	"Initialize session-based encryption. (delayed change)"
	session [object!]
	encoding [integer!]
	auth "Pass hash"
][
	if not binary? auth [exit]
	; Generate session key based on user's password hash + server key + challenge:
	make-session-key session auth
	; Open the encryption ports if necessary:
	session/encode: encoding
	close-crypt-ports
	if 'aes = key-type? session/encode [
		open-crypt-ports session
	]
]

confirm-session: func [
	"Confirm that the user actually authenticated ok. (delayed change)"
	session [object!]
][
	session/login-date: now
	; log-user session/user-id user-name remote-ip ??
]

change-session: func [
	"Handles changes to session after last reply."
	session [object!]
	/local new-changes
][
	; This function is used to make changes to the session after the
	; previous reply was made. This is normally required for changes
	; to encryption, because the client must be told (via a reply)
	; that the encryption is about to change (e.g. to gen a new key).
	new-changes: copy []
	foreach item session/changed [
		switch item [
			login  [
				crypt-session session session/challenge/3 auth-user? session/user-id
				append new-changes 'confirm
			]
			reset [reset-session session]
			confirm [
				confirm-session session
			]
		]
	]
	insert clear session/changed new-changes
]
